﻿using System;
using System.Threading;
using System.Runtime.Serialization;
using Wintellect;


internal static class CoolTypes {
   public static void Demos() {
      GCBeepDemo();
      DisposerDemo();
      ExceptionDemo();
      SingletonDemo();
      InitOnceDemo();
      StructureEnumsDemo();
   }

   #region GCBeep
   private static void GCBeepDemo() {
      // Constructing a single GCBeep object causes a beep to
      // occur every time a garbage collection starts.
      new GCBeep();

      // Construct a lot of objects.
      for (Int32 x = 0; x < 10000; x++) {
         Console.WriteLine(x);
         Byte[] b = new Byte[1000];
      }
      GCBeep.Silence = true;
   }

   #endregion

   #region Disposer
   private static void DisposerDemo() {
      using (new Disposer(delegate { Console.WriteLine("Operation is complete."); })) {
         for (Int32 x = 0; x < 50; x++) {
            Console.WriteLine(x);
            Thread.Sleep(50);
         }
      }
   }
   #endregion

   #region Exception<T>
   [Serializable]
   private sealed class DiskFullExceptionArgs : ExceptionArgs {
      private String s_volume;

      public DiskFullExceptionArgs(String volume) { s_volume = volume; }

      public String Volume { get { return s_volume; } }

      public override String ToString() { return "Volume=" + s_volume; }
   }

   private static void ExceptionDemo() {
      try {
         DiskFullExceptionArgs dfea = new DiskFullExceptionArgs("C");
         throw new Exception<DiskFullExceptionArgs>(dfea, "Disk full");
      }
      catch (Exception<DiskFullExceptionArgs> e) {
         Console.WriteLine("Caught exception, Type={0}", e.GetType());
         Console.WriteLine("Message={0}", e.Message);
         Console.WriteLine("Volume={0}", e.Args.Volume);
      }
   }
   #endregion

   private sealed class Foo { }
   private static void SingletonDemo() {
      Singleton<Foo>.Creator creator = delegate { Console.WriteLine("Creating singleton."); return new Foo(); };

      Singleton<Foo> theFoo = new Singleton<Foo>(SingletonRaceLoser.Block, creator);
      Foo foo = theFoo.Value;
      foo = theFoo.Value;

      theFoo = new Singleton<Foo>(SingletonRaceLoser.GC, creator);
      foo = theFoo.Value;
      foo = theFoo.Value;
   }

   private static void InitOnceDemo() {
      WaitCallback initFunc = delegate { Console.WriteLine("Initializing"); };

      InitializeOnce io = new InitializeOnce();
      io.Initialize(initFunc, null);
      io.Initialize(initFunc, null);
   }

   private struct Fruit {
      private Int32 m_value;
      public static readonly Fruit Orange = 1;
      public static readonly Fruit Grape = 1;
      public static readonly Fruit Banana = 2;
      public static implicit operator Fruit(Int32 value) { return new Fruit(value); }
      private Fruit(Int32 value) { m_value = value; }
      public Boolean IsYellow() { return m_value == Banana; }
      public static Boolean operator ==(Fruit f1, Fruit f2) { return f1.m_value == f2.m_value; }
      public static Boolean operator !=(Fruit f1, Fruit f2) { return !(f1 == f2); }
      public override Boolean Equals(object obj) { return this == (Fruit)obj; }
      public override int GetHashCode() { return m_value.GetHashCode(); }
   }
   private static void StructureEnumsDemo() {
      Fruit f = Fruit.Banana;
      Console.WriteLine("{0} is{1} yellow", f, f.IsYellow() ? null : " not");
   }

   #region Enum Inheritance
   public sealed class BaseEnum {
      private BaseEnum() { }

      private static BaseEnum s_singleton = new BaseEnum();

      public static BaseEnum Rights {
         get { return s_singleton; }
      }

      public readonly UInt32 None = 0x00000000;
      public readonly UInt32 Delete = 0x00010000;
      public readonly UInt32 ReadControl = 0x00020000;
   }

   public struct DerivedEnum {
      private DerivedEnum(UInt32 rights) {
         m_processRights = rights;
      }

      public static implicit operator DerivedEnum(UInt32 rights) {
         return new DerivedEnum(rights);
      }

      public static explicit operator UInt32(DerivedEnum rights) {
         return rights.m_processRights;
      }

      public static DerivedEnum operator|(DerivedEnum lhs, DerivedEnum rhs) {
         return new DerivedEnum(lhs.m_processRights | rhs.m_processRights);
      }

      public static BaseEnum Generic { get { return BaseEnum.Rights; } }

      public static readonly DerivedEnum Terminate = 0x00000001;
      public static readonly DerivedEnum CreateThread = 0x00000002;
      private UInt32 m_processRights;
   }

   public struct TestEnumInheritance {
      static void Bar(DerivedEnum rights) {
      }

      static void Blah() {
         Bar(DerivedEnum.CreateThread | DerivedEnum.Generic.Delete);
      }
   }
   #endregion
}


#if false
internal static class CSharpLanguageConstructs {
   public static void Demos() {
      //CheckedAndUncheckedCode();
      CastOperators();
      Indexers();
   }

   private static void CheckedAndUncheckedCode() {
      Int32 val = 2;
      try {
         while (true) { checked { val *= val; } };
      }
      catch (OverflowException) {
      }

      Int32 n = -1;
      UInt32 m = unchecked((UInt32)n);

      Decimal d = 0;
      unchecked {
         try {
            while (true) d *= d;
         }
         catch (OverflowException) {
         }
      }
   }

   private static void CastOperators() {
      Foo f = 0;
      Console.WriteLine(f);
      Int32 dayOfYear = (Int32)f;
      Console.WriteLine(dayOfYear);
   }

   private sealed class Foo {
      DateTime m_dateTime;
      public Foo(Int32 dayOfYear) {
         m_dateTime = this[dayOfYear];
      }
      public static implicit operator Foo(Int32 dayOfYear) {
         return new Foo(dayOfYear);
      }
      public static explicit operator Int32(Foo foo) {
         return foo.m_dateTime.DayOfYear;
      }
      public DateTime this[Int32 dayOfYear] {
         get { return DateTime.Now.Add(new TimeSpan(dayOfYear, 0, 0, 0)); }
      }
      public override string ToString() {
         return m_dateTime.ToString();
      }
   }
   private static void Indexers() {
      for (Int32 dayOfYear = 0; dayOfYear < 365; dayOfYear += 30)
         Console.WriteLine(new Foo(0)[dayOfYear]);
   }
}
#endif