/******************************************************************************
Module:  PerfTips.cs
Notices: Copyright (c) 2008 Jeffrey Richter
******************************************************************************/


using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Reflection;
using System.Runtime.CompilerServices;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using Wintellect;
using System.Linq;


///////////////////////////////////////////////////////////////////////////////


public static class Program {
   [MethodImpl(MethodImplOptions.NoInlining)]
   [DebuggerStepThrough]
   private static void TestName() {
      StackFrame[] frames = new StackTrace().GetFrames();
      StringBuilder testName = new StringBuilder("****************************************\n");
      testName.Append(frames[1].GetMethod().Name);
      testName.Replace('_', ':');
      for (Int32 index = 1; index < testName.Length; index++) {
         // If this is an uppercase character, insert a space before it
         if (Char.IsUpper(testName[index])) {
            testName.Insert(index, ' ');
            index++;
         }
      }
      testName.Append("\n****************************************");
      TestName(testName.ToString());
   }

   private static void TestName(String testName) {
      Console.WriteLine();
      Console.WriteLine(testName);
   }

   public static void Main() {
      // Pre-empt a lot of other apps for more accurate results
      Process.GetCurrentProcess().PriorityClass = ProcessPriorityClass.High;
      Thread.CurrentThread.Priority = ThreadPriority.Highest;

      // Force the CodeTimer.Time to be jitted/loaded/whatever.
      CodeTimer.Time(String.Empty, 1, () => { Thread.Sleep(0); });

      //CodeTimer.Time ( "Sleep vs. Cycle Times 5000" , 1 , ( ) => { Thread.Sleep ( 5000 ); } );
      //CodeTimer.Time ( "Sleep vs. Cycle Times 15000" , 1 , ( ) => { Thread.Sleep ( 15000 ); } );

      StringPerformanceTests();

      VariableAccessPerformanceTests();

      MethodAndReflectionPerformanceTests();

      CollectionPerformanceTests();

      ThreadingPerformanceTests();

      Console.WriteLine("\nDone All Performance Tests");
   }

   #region String Performance Tests
   private static void StringPerformanceTests() {
      TestName();
      const Int32 iterations = 100 * 1000;

      TestName("String Concatenation Tests");

      String s = String.Empty;
      CodeTimer.Time("Concatenating to a String", iterations,
         () => s += "X");

      StringBuilder sb = new StringBuilder(String.Empty);
      CodeTimer.Time("Concatenating to a StringBuilder", iterations,
         () => sb.Append("X"));

      TestName("String Interning Tests");

      String[] theTypes = TypesFromAssembly();
      String word = "NonExistantType";
      CodeTimer.Time("Lookup without string interning", iterations,
         () => {
            for (Int32 i = 0; i < theTypes.Length; i++) {
               if (String.Equals(word, theTypes[i], StringComparison.Ordinal))
                  break;
            }
         });


      CodeTimer.Time("Time to intern string array", 1,
         () => {
            word = String.Intern(word);
            for (int i = 0; i < theTypes.Length; i++)
               theTypes[i] = String.Intern(theTypes[i]);
         });

      CodeTimer.Time("Lookup with string interning", iterations,
         () => {
            for (Int32 i = 0; i < theTypes.Length; i++)
               if (Object.ReferenceEquals(word, theTypes[i]))
                  break;
         });

      TestName("String InterOp Tests");

      CodeTimer.Time("Interop with ANSI String", iterations,
         () => lstrlenAnsi("This is a string"));

      CodeTimer.Time("Interop with UNICODE String", iterations,
         () => lstrlenUnicode("This is a string"));
   }

   private static String[] TypesFromAssembly() {
      return (typeof(Object).Assembly.GetTypes().Select(T => T.FullName).ToArray());
   }

   [DllImport("Kernel32", ExactSpelling = true, EntryPoint = "lstrlenA")]
   private static extern Int32 lstrlenAnsi(String s);

   [DllImport("Kernel32", ExactSpelling = true, EntryPoint = "lstrlenW", CharSet = CharSet.Unicode)]
   private static extern Int32 lstrlenUnicode(String s);

   #endregion

   #region Variable Access Performance Tests
   private static void VariableAccessPerformanceTests() {
      TestName();

      Int32 iterations = 100 * 1000 * 1000;

      TestName("Checked and Unchecked Numeric Operation Tests");

      int i = 0;
      CodeTimer.Time("Checked operations", iterations,
         () => i = checked(i + 1));

      i = 0;
      CodeTimer.Time("Unchecked operations", iterations,
         () => i = unchecked(i + 1));

      iterations = 500000000;
      FieldAccess fa = new FieldAccess();

      TestName("Static Reference Field Access Tests");

      Object o = null;
      CodeTimer.Time("Read  static reference field", iterations,
         () => o = FieldAccess.s_refType);

      CodeTimer.Time("Write static reference field", iterations,
         () => FieldAccess.s_refType = o);

      TestName("Static Value Field Access Tests");

      IntPtr p = IntPtr.Zero;
      CodeTimer.Time("Read  static value field", iterations,
         () => p = FieldAccess.s_valType);

      CodeTimer.Time("Write static value field", iterations,
         () => FieldAccess.s_valType = p);

      TestName("Instance Reference Field Access Tests");

      CodeTimer.Time("Read  instance reference field", iterations,
         () => o = fa.m_refType);

      CodeTimer.Time("Write instance reference field", iterations,
         () => fa.m_refType = o);

      TestName("Instance Value Field Access Tests");

      CodeTimer.Time("Read  instance value field", iterations,
         () => p = fa.m_valType);

      CodeTimer.Time("Write instance value field", iterations,
         () => fa.m_valType = p);

      TestName("Instance Field Normal Object vs. MarshalByRefObject Tests");

      iterations = 100000000;
      NormalObject no = new NormalObject();
      CodeTimer.Time("Normal object instance field", iterations,
         () => no.m_field++);

      MBRObject mbro = new MBRObject();
      CodeTimer.Time("MarshalByRefObject instance field", iterations,
         () => mbro.m_field++);

      TestName("Nullable vs. Non-nullable Integer Tests");

      iterations = 100000000;
      Int32 integer = 0;
      CodeTimer.Time("Non-nullable variable", iterations,
         () => integer++);

      Int32? nullableInteger = 0;
      CodeTimer.Time("Nullable variable", iterations,
         () => nullableInteger++);
   }

   private sealed class FieldAccess {
      public static IntPtr s_valType; // Value type static field
      public static Object s_refType; // Reference type static field

      public IntPtr m_valType = IntPtr.Zero; // Value type instance field
      public Object m_refType = null;        // Reference type instance field
   }

   private sealed class NormalObject : Object { public Int32 m_field = 0; }
   private sealed class MBRObject : MarshalByRefObject { public Int32 m_field = 0; }

   #endregion

   #region Method and Reflection Performance Tests

   private static void MethodAndReflectionPerformanceTests() {
      TestName();
      Int32 iterations = 500 * 1000 * 1000;

      TestName("Method Calling Tests");

      CodeTimer.Time("Calling static", iterations,
         () => Methods.StaticMethod());

      Methods methods = new Methods();
      IInterface iface = methods;

      CodeTimer.Time("Calling instance", iterations,
         () => methods.InstanceMethod());

      CodeTimer.Time("Calling virtual", iterations,
         () => methods.VirtualMethod());

      CodeTimer.Time("Calling interface", iterations,
         () => iface.InterfaceMethod());

      DelegateMethod dm = new DelegateMethod(Methods.DelegateStaticMethod);
      CodeTimer.Time("Calling static via delegate", iterations,
         () => dm());

      dm = new DelegateMethod(methods.DelegateInstanceMethod);
      CodeTimer.Time("Calling instance via delegate", iterations,
         () => dm());

      CodeTimer.Time("Calling static generic", iterations,
         () => Methods.StaticGenericMethod<Object>());

      CodeTimer.Time("Calling instance generic", iterations,
         () => methods.InstanceGenericMethod<Object>());

      CodeTimer.Time("Calling via Reflection, bind and Invoke (time x 1000)", iterations / 1000,
         () => typeof(Methods).GetMethod("ReflectionMethod").Invoke(null, null));

      MethodInfo mi = typeof(Methods).GetMethod("ReflectionMethod");
      CodeTimer.Time("Calling via reflection Invoke (time x 1000)", iterations / 1000,
         () => mi.Invoke(null, null));

      dm = (DelegateMethod)
              Delegate.CreateDelegate(typeof(DelegateMethod), mi);
      CodeTimer.Time("Calling reflection via delegate", iterations, () => dm());

      TestName("Methods with Variable Number of Arguments Tests");

      iterations = 50000000;

      CodeTimer.Time("0 arg direct", iterations, () => M0A());

      CodeTimer.Time("0 arg variable", iterations, () => MVA());

      Console.WriteLine();

      CodeTimer.Time("1 arg direct", iterations, () => M1A(null));

      // NOTE: Clever C# compiler optimization: if null is passed as the 
      // argument to a params parameter, the compiler passes null without 
      // new'ing up the array and initializing the element to null.
      CodeTimer.Time("1 arg variable", iterations, () => MVA(null));

      Console.WriteLine();

      CodeTimer.Time("2 arg direct", iterations, () => M2A(null, null));

      CodeTimer.Time("2 arg variable", iterations, () => MVA(null, null));

      Console.WriteLine();

      CodeTimer.Time("3 arg direct", iterations, () => M3A(null, null, null));

      CodeTimer.Time("3 arg variable", iterations, () => MVA(null, null, null));

      Console.WriteLine();

      CodeTimer.Time("4 arg direct", iterations, () => M4A(null, null, null, null));

      CodeTimer.Time("4 arg variable", iterations, () => MVA(null, null, null, null));

      Console.WriteLine();

      CodeTimer.Time("5 arg direct", iterations, () => M5A(null, null, null, null, null));

      CodeTimer.Time("5 arg variable", iterations, () => MVA(null, null, null, null, null));
   }

   [MethodImpl(MethodImplOptions.NoInlining)]
   private static void M0A() { }

   [MethodImpl(MethodImplOptions.NoInlining)]
   private static void M1A(Object o1) { }

   [MethodImpl(MethodImplOptions.NoInlining)]
   private static void M2A(Object o1, Object o2) { }

   [MethodImpl(MethodImplOptions.NoInlining)]
   private static void M3A(Object o1, Object o2, Object o3) { }

   [MethodImpl(MethodImplOptions.NoInlining)]
   private static void M4A(Object o1, Object o2, Object o3, Object o4) { }

   [MethodImpl(MethodImplOptions.NoInlining)]
   private static void M5A(Object o1, Object o2, Object o3, Object o4, Object o5) { }

   [MethodImpl(MethodImplOptions.NoInlining)]
   private static void MVA(params Object[] os) { }

   private interface IInterface {
      void InterfaceMethod();
   }

   private class Methods : IInterface {
      [MethodImpl(MethodImplOptions.NoInlining)]
      public static void StaticMethod() { }

      [MethodImpl(MethodImplOptions.NoInlining)]
      public void InstanceMethod() { }

      [MethodImpl(MethodImplOptions.NoInlining)]
      public virtual void VirtualMethod() { }

      [MethodImpl(MethodImplOptions.NoInlining)]
      public void InterfaceMethod() { }

      [MethodImpl(MethodImplOptions.NoInlining)]
      public static void DelegateStaticMethod() { }

      [MethodImpl(MethodImplOptions.NoInlining)]
      public void DelegateInstanceMethod() { }

      [MethodImpl(MethodImplOptions.NoInlining)]
      public static void StaticGenericMethod<T>() { }

      [MethodImpl(MethodImplOptions.NoInlining)]
      public void InstanceGenericMethod<T>() { }

      [MethodImpl(MethodImplOptions.NoInlining)]
      public static void ReflectionMethod() { }
   }

   private delegate void DelegateMethod();

   #endregion

   #region Collection Performance Tests

   private static void CollectionPerformanceTests() {
      Int32 iterations = 5000;
      // We are keeping the one and two dimensional arrays exactly the same 
      // size in number of total elements.
      const Int32 numElements = 250;
      Int32 sum = 0;

      TestName();

      TestName("Reading and writing reference vs. value type arrays");

      IntPtr dummy = IntPtr.Zero;
      IntPtr[] aValues = new IntPtr[numElements * numElements];
      CodeTimer.Time("Reading array of values", iterations,
         () => {
            for (int i = 0; i < aValues.Length; i++)
               dummy = aValues[i];
         });

      CodeTimer.Time("Writing array of references", iterations,
         () => {
            for (int i = 0; i < aValues.Length; i++)
               aValues[i] = dummy;
         });

      Object o = null;
      Object[] aRefs = new Object[numElements * numElements];
      CodeTimer.Time("Reading array of references", iterations,
         () => {
            for (int i = 0; i < aRefs.Length; i++)
               o = aRefs[i];
         });

      o = null;
      CodeTimer.Time("Writing array of references", iterations,
         () => {
            for (int i = 0; i < aRefs.Length; i++)
               aRefs[i] = o;
         });

      TestName("Zero-based one, two, and jagged array access tests");

      Int32[] aint = new Int32[numElements * numElements];
      CodeTimer.Time("1-dim zero-based access: " + aint.GetType(), iterations,
         () => {
            for (Int32 e = 0; e < aint.Length; e++)
               sum += aint[e];
         });

      Int32[,] a = new Int32[numElements, numElements];
      CodeTimer.Time("2-dim zero-based access: " + a.GetType(), iterations,
         () => {
            for (Int32 e = 0; e < numElements; e++)
               for (Int32 f = 0; f < numElements; f++)
                  sum += a[e, f];
         });

      Int32[][] aJagged = new Int32[numElements][];
      for (Int32 i = 0; i < numElements; i++)
         aJagged[i] = new Int32[numElements];
      CodeTimer.Time("2-dim zero-based jagged access: " + aJagged.GetType(), iterations,
         () => {
            for (Int32 e = 0; e < numElements; e++)
               for (Int32 f = 0; f < numElements; f++)
                  sum += aJagged[e][f];
         });

      TestName("CreateInstance of zero and one-based two dim array tests");

      // 2-Dim array access: CreateInstance with start index of [0, 0]
      Int32[,] a2 = (Int32[,]) Array.CreateInstance(typeof(Int32), numElements, numElements);
      CodeTimer.Time("2-dim zero-base CreateInstance'd access: " + a2.GetType(), iterations,
         () => {
            for (Int32 e = 0; e < numElements; e++)
               for (Int32 f = 0; f < numElements; f++)
                  sum += a2[e, f];
         });

      // 2-Dim array access: CreateInstance with start index of [1, 1]
      Array a3 = Array.CreateInstance(typeof(Int32), new Int32[] { numElements, numElements }, new Int32[] { 1, 1 });
      CodeTimer.Time("2-dim one-based CreateInstance'd GetValue access (x1000): " + a3.GetType(), iterations / 1000,
         () => {
            for (Int32 e = 1; e <= numElements; e++)
               for (Int32 f = 1; f <= numElements; f++)
                  sum += (Int32) a3.GetValue(e, f);
         });

      // 1-Dim array access: CreateInstance with start index of [0].
      // This is to show that you can cast to an Int32[] without problems.
      Int32[] a4 = (Int32[]) Array.CreateInstance(typeof(Int32), numElements * numElements);
      try {
         // Here we want to show you can create a non-zero based array, but 
         // cannot cast and there's no good C# syntax to describe it.
         Int32[] a5 = (Int32[]) Array.CreateInstance(typeof(Int32), new Int32[] { numElements * numElements }, new Int32[] { 1 });
      }
      catch (InvalidCastException ex) {
         Console.WriteLine("\n" + ex.Message);
      }

      TestName("One dim CreateInstance array one-based test");

      // Here we have to use the Array type (the bad syntax) and our perf is 
      // bad as we have have to call GetValue/SetValue to access elements.
      Array a6 = Array.CreateInstance(typeof(Int32), new Int32[] { numElements * numElements }, new Int32[] { 1 });
      CodeTimer.Time("1-dim non-0 based CreateInstance'd GetValue access (x100): " + a6.GetType(), iterations / 100,
         () => {
            for (Int32 e = 1; e <= a6.Length; e++)
               sum += (Int32) a6.GetValue(e);
         });


      TestName("Two dim CreateInstance array one-based test");

      // Interestingly, you can use natural C# array syntax to access multi dimensional
      // arrays.
      Int32[,] a7 = (Int32[,]) Array.CreateInstance(typeof(Int32),
                                        new Int32[] { numElements, numElements },
                                        new Int32[] { 1, 1 });
      CodeTimer.Time("2-dim non-zero based CreateInstance'd access " + a7.GetType(), iterations,
         () => {
            for (Int32 e = 1; e < numElements; e++)
               for (Int32 f = 1; f < numElements; f++)
                  sum += a7[e, f];
         });


      TestName("Column vs. Row Access Tests");

      Int32[,] a8 = new Int32[numElements, numElements];
      CodeTimer.Time("Column access test", iterations,
         () => {
            for (int col = 0; col < numElements; col++)
               for (int row = 0; row < numElements; row++)
                  a8[row, col] = 6;
         });

      CodeTimer.Time("Row access test", iterations,
         () => {
            for (int row = 0; row < numElements; row++)
               for (int col = 0; col < numElements; col++)
                  a8[row, col] = 6;
         });


      TestName("Growing vs. Presizing Collections Tests");

      iterations = 10000000;

      List<Object> lo1 = new List<Object>();
      CodeTimer.Time("On-demand collection growth", iterations,
         () => lo1.Add(null));

      List<Object> lo2 = new List<Object>(iterations);
      CodeTimer.Time("Pre-sized collection", iterations,
         () => lo2.Add(null));


      TestName("Adding One Element at a Time vs AddRange Collections Tests");

      Object[] oa = new Object[iterations];

      List<Object> lo3 = new List<Object>();
      CodeTimer.Time("One element at a time", 1,
         () => {
            for (int i = 0; i < iterations; i++)
               lo3.Add(oa[i]);
         }); ;

      List<Object> lo4 = new List<Object>();
      CodeTimer.Time("Add range all at once", 1,
         () => lo4.AddRange(oa));

      TestName("foreach vs. for vs. ForEach Tests");

      iterations = 100 * 1000 * 1000;
      List<Object> l5 = new List<Object>(iterations);
      for (Int32 x = 0; x < iterations; x++)
         l5.Add(null);

      CodeTimer.Time("Using C# foreach statement", 1,
         () => {
            foreach (Object o2 in l5)
               M(o2);
         });

      CodeTimer.Time("Using C# foreach statement with cast to IEnumerable", 1,
         () => {
            foreach (Object o3 in (IEnumerable<Object>) l5)
               M(o3);
         });

      CodeTimer.Time("Using C# for statement", 1,
         () => {
            for (Int32 x = 0; x < iterations; x++)
               M(l5[x]);
         });

      CodeTimer.Time("Using Lists's ForEach method", 1,
         () => l5.ForEach(M));
   }

   [MethodImpl(MethodImplOptions.NoInlining)]
   private static void M(Object o) { }
   #endregion

   #region Threading Performance Tests

   private static void ThreadingPerformanceTests() {
      TestName();
      const Int32 iterations = 1000000;

      TestName("Lock vs. ReaderWriterLock vs. ReaderWriterLockSlim vs. Handle-based");

      Object o = new Object();
      CodeTimer.Time("Lock perf: Monitor", iterations,
         () => {
            Monitor.Enter(o);
            Monitor.Exit(o);
         });

      ReaderWriterLockSlim rwls = new ReaderWriterLockSlim();
      CodeTimer.Time("Lock perf: ReaderWriterLockSlim", iterations,
         () => {
            rwls.EnterWriteLock();
            rwls.ExitWriteLock();
         });

      ReaderWriterLock rwl = new ReaderWriterLock();
      CodeTimer.Time("Lock perf: ReaderWriterLock", iterations,
         () => {
            rwl.AcquireWriterLock(Timeout.Infinite);
            rwl.ReleaseWriterLock();
         });

      Mutex mutex = new Mutex();
      CodeTimer.Time("Lock perf: Mutex", iterations,
         () => {
            mutex.WaitOne();
            mutex.ReleaseMutex();
         });
   }
   #endregion
}